import datetime
import os

from chainerui.database import db
from chainerui.models.result import Result


def _list_result_paths(target_path, log_file_name='log'):
    """list_result_paths."""

    result_list = []

    for root, _dirs, _files in os.walk(os.path.abspath(target_path)):
        for name in _files:
            if name == log_file_name:
                result_list.append(root)

    return result_list


def _register_result(project_id, result_path):
    result_path = os.path.abspath(result_path)

    contain_log_file = os.path.isfile(os.path.join(result_path, 'log'))

    if not contain_log_file:
        return

    Result.create(project_id=project_id, path_name=result_path)


def collect_results(project, force=False):
    """collect_results."""
    if not project.crawlable:
        return project

    now = datetime.datetime.now()

    if (now - project.updated_at).total_seconds() < 4 and (not force):
        return project

    result_paths = []
    if os.path.isdir(project.path_name):
        result_paths.extend(_list_result_paths(project.path_name))

    registered_results = db.session.query(Result.path_name).filter_by(
        project_id=project.id
    ).all()
    registered_paths = {r.path_name for r in registered_results}

    for result_path in result_paths:
        if result_path not in registered_paths:
            _register_result(project.id, result_path)

    project.updated_at = datetime.datetime.now()

    db.session.commit()

    return project



"""empty message

Revision ID: c29d1b4b67fe
Revises: 6cd7f193a3de
Create Date: 2018-08-28 09:54:01.296143

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = 'c29d1b4b67fe'
down_revision = '6cd7f193a3de'
branch_labels = None
depends_on = None


def upgrade():
    op.create_index('idx_result_project_id', 'result', ['project_id'])
    op.create_index('idx_log_result_id', 'log', ['result_id'])


def downgrade():
    op.drop_index('idx_log_result_id')
    op.drop_index('idx_result_project_id')



"""empty message.

Revision ID: 6cc68bf684ac
Revises: 1104edb51c0d
Create Date: 2017-09-20 09:52:16.837947

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '6cc68bf684ac'
down_revision = '1104edb51c0d'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('result') as batch_op:
        batch_op.add_column(
            sa.Column('project_id', sa.Integer(), nullable=True))
        sa.ForeignKeyConstraint(['project_id'], ['project.id'])


def downgrade():
    with op.batch_alter_table('result') as batch_op:
        batch_op.drop_column('project_id')



import contextlib
import datetime
import json
import os
import warnings

import filelock

from chainerui.logging import logger


CHAINERUI_ASSETS_METAFILE_NAME = '.chainerui_assets'


class _Summary(object):

    def __init__(self):
        self.cache = []
        self.filename = CHAINERUI_ASSETS_METAFILE_NAME
        self.default_output_path = ''
        self.out = None
        self.saved_idx = 0

    def add(self, value):
        self.cache.append(value)

    def get_outpath(self, out):
        # return output path led by argument > global priority
        if out is not None:
            return out

        if self.out is None:
            warnings.warn('Output directory is not set, so empty path(\'\') '
                          'is set as output directory')
            return self.default_output_path
        return self.out

    def save(self, out, timeout):
        filepath = os.path.join(out, self.filename)
        lockpath = filepath + '.lock'

        try:
            with filelock.FileLock(lockpath, timeout=timeout):
                saved_assets_list = []
                if os.path.isfile(filepath):
                    with open(filepath) as f:
                        saved_assets_list = json.load(f)
                saved_assets_list.extend(self.cache[self.saved_idx:])
                with open(filepath, 'w') as f:
                    json.dump(saved_assets_list, f, indent=4)
                self.saved_idx = len(self.cache)
        except filelock.Timeout:
            logger.error('Process to write a list of assets is timeout')


_chainerui_asset_observer = _Summary()


class _Reporter(object):

    def __init__(self, prefix=None, out=None, subdir='', **kwargs):
        self.prefix = prefix
        self.out = _chainerui_asset_observer.get_outpath(out)
        self.subdir = subdir
        self.value = kwargs

        self.images = {}
        self.audios = {}
        self.texts = {}
        self.count = 0

    def image(self, images, name=None, ch_axis=1, row=0, mode=None,
              batched=True, subdir=''):
        """Summarize images to visualize.

        Args:
            images (:class:`numpy.ndarray` or :class:`cupy.ndarray` or \
                :class:`chainer.Variable`): batch of images. If Number of
                dimension is 3 (or 2 when set `batched=False`), the pixels
                assume as black and white image.
            name (str): name of image. set as column name. when not setting,
                assigned ``'image'`` + sequential number.
            ch_axis (int): index number of channel dimension. set 1 by default.
                if the images don't have channel axis, this parameter is
                ignored.
            row (int): row size to visualize batched images. when set 0,
                show on unstuck. if images set only one image, the row size
                will be ignored.
            mode (str): if the images are not RGB or RGBA space, set their
                color space code. ChainerUI supports 'HSV'.
            batched (bool): if the image is not batched, set ``False``.
            subdir (str): sub-directory path of output.
        """
        from chainerui.report.image_report import check_available
        if not check_available():
            return
        from chainerui.report.image_report import report as _image

        col_name = self.get_col_name(name, 'image')
        out_dir, rel_out_dir = self.get_subdir(subdir)
        filename, _ = _image(
            images, out_dir, col_name, ch_axis, row, mode, batched)
        self.images[col_name] = os.path.join(rel_out_dir, filename)

        self.count += 1

    def audio(self, audio, sample_rate, name=None, subdir=''):
        """Summarize audio to listen on web browser.

        Args:
            audio (:class:`numpy.ndarray` or :class:`cupy.ndarray` or \
                :class:`chainer.Variable`): sampled wave array.
            sample_rate (int): sampling rate.
            name (str): name of image. set as column name. when not setting,
                assigned ``'audio'`` + sequential number.
            subdir (str): sub-directory path of output.
        """

        from chainerui.report.audio_report import check_available
        if not check_available():
            return
        from chainerui.report.audio_report import report as _audio

        col_name = self.get_col_name(name, 'audio')
        out_dir, rel_out_dir = self.get_subdir(subdir)
        filename, _ = _audio(audio, sample_rate, out_dir, col_name)
        self.audios[col_name] = os.path.join(rel_out_dir, filename)

        self.count += 1

    def text(self, text, name=None):
        """Summarize texts to show on a browser.

        Arguments:
            text (str): generated text.
            name (str): name of text. set as column name. when not setting,
                assigned ``'text'``.
        """
        col_name = self.get_col_name(name, 'text')
        self.texts[col_name] = text

        self.count += 1

    def get_col_name(self, name, media_type):
        col_name = name
        if col_name is None:
            col_name = '{}_{:d}'.format(media_type, self.count)
        if self.prefix is not None:
            col_name = self.prefix + col_name
        return col_name

    def get_subdir(self, subdir):
        rel_out_dir = os.path.join(self.subdir, subdir)
        out_dir = os.path.join(self.out, rel_out_dir)
        if not os.path.isdir(out_dir):
            os.makedirs(out_dir)
        return out_dir, rel_out_dir

    def save(self, timeout):
        cached = False
        if self.images:
            self.value['images'] = self.images
            cached = True
        if self.audios:
            self.value['audios'] = self.audios
            cached = True
        if self.texts:
            self.value['texts'] = self.texts
            cached = True
        if not cached:
            return
        self.value['timestamp'] = datetime.datetime.now().isoformat()
        _chainerui_asset_observer.add(self.value)
        _chainerui_asset_observer.save(self.out, timeout)


def set_out(path):
    """Set output path.

    Summary module requires output directory. Once set output path using this
    function, summary module shares the path.

    Args:
        path (str): directory path of output.
    """

    _chainerui_asset_observer.out = path


@contextlib.contextmanager
def reporter(prefix=None, out=None, subdir='', timeout=5, **kwargs):
    """Summarize media assets to visualize.

    ``reporter`` function collects media assets by the ``with`` statement and
    aggregates in same row to visualize. This function returns an object which
    provides the following methods.

    * :meth:`~chainerui.summary._Reporter.image`: collect images. almost same \
        as :func:`~chainerui.summary.image`
    * :meth:`~chainerui.summary._Reporter.audio`: collect audio. almost same \
        as :func:`~chainerui.summary.audio`
    * :meth:`~chainerui.summary._Reporter.text`: collect text. almost same \
        as :func:`~chainerui.summary.text`

    Example of how to set several assets::

       >>> from chainerui.summary import reporter
       >>> summary.set_out('/path/to/output')  # same as 'log' file directory
       >>>
       >>> with reporter(epoch=1, iteration=10) as r:
       >>>     r.image(image_array1)
       >>>     r.image(image_array2)
       >>>     r.audio(audio_array, 44100)
       >>> # image_array1 and image_array2 are visualized on a browser
       >>> # audio_array can be listened on a browser

    Args:
        prefix (str): prefix of column name.
        out (str): directory path of output.
        subdir (str): sub-directory path of output.
        **kwargs (dict): key-value pair to show as description. regardless of
            empty or not, timestamp is added.
    """

    report = _Reporter(prefix, out, subdir, **kwargs)
    yield report
    report.save(timeout)


def image(images, name=None, ch_axis=1, row=0, mode=None, batched=True,
          out=None, subdir='', timeout=5, **kwargs):
    """Summarize images to visualize.

    Array of images are converted as image format (PNG format on default),
    saved to output directory, and reported to the ChainerUI server.
    The images are saved every called this function. The images will be shown
    on `assets` endpoint vertically. If need to aggregate images in a row, use
    :func:`~chainerui.summary.reporter`.

    Examples of how to set arguments::

       >>> from chainerui import summary
       >>> summary.set_out('/path/to/log')  # same as 'log' file directory
       >>>
       >>> x.shape  # = [Batchsize, Channel, Height, Width]
       (10, 3, 5, 5)
       >>> summary.image(x, name='test')  # images are tiled as 1x10
       >>> summary.image(x, name='test', row=2)  # images are tiled as 2x5
       >>>
       >>> x.shape  # = [C, H, W]
       (3, 5, 5)
       >>> # need to set as a non-batched image and channel axis explicitly
       >>> summary.image(x, name='test', ch_axis=0, batched=False)
       >>>
       >>> x.shape  # = [B, H, W, C]
       (10, 5, 5, 3)
       >>> # need to set channel axis explicitly
       >>> summary.image(x, name='test', ch_axis=-1, row=2)
       >>>
       >>> x.shape  # = [H, W, C]
       (5, 5, 3)
       >>> # need to set as a non-batched image
       >>> summary.image(x, name='test', ch_axis=-1, batched=False)
       >>>
       >>> x.shape  # = [B, H, W], grayscale images
       (10, 5, 5)
       >>> summary.image(x, name='test')  # image are tiled as 1x10
       >>> summary.image(x, name='test', row=2)  # image are tiled as 2x5
       >>>
       >>> x.shape  # = [H, W], a grayscale image
       (5, 5)
       >>> # need to set as a non-bathed image
       >>> summary.image(x, name='test', batched=False)

    Add description about the image::

       >>> summary.image(x, name='test', epoch=1, iteration=100)
       >>> # 'epoch' and 'iteration' column will be shown.

    Args:
        images (:class:`numpy.ndarray` or :class:`cupy.ndarray` or \
            :class:`chainer.Variable`): batch of images. If Number of dimension
            is 3 (or 2 when set `batched=False`), the pixels assume as
            black and white image.
        name (str): name of image. set as column name. when not setting,
            assigned ``'image'``.
        ch_axis (int): index number of channel dimension. set 1 by default.
            if the images don't have channel axis, this parameter is ignored.
        row (int): row size to visualize batched images. when set 0,
            show on unstuck. if images set only one image, the row size
            will be ignored.
        mode (str): if the images are not RGB or RGBA space, set their
            color space code. ChainerUI supports 'HSV'.
        batched (bool): if the image is not batched, set ``False``.
        out (str): directory path of output.
        subdir (str): sub-directory path of output.
        **kwargs (dict): key-value pair to show as description. regardless of
            empty or not, timestamp on created the image is added.
    """

    from chainerui.report.image_report import check_available
    if not check_available():
        return
    from chainerui.report.image_report import report as _image

    out_root = _chainerui_asset_observer.get_outpath(out)
    out_path = os.path.join(out_root, subdir)
    if not os.path.isdir(out_path):
        os.makedirs(out_path)
    col_name = name
    if col_name is None:
        col_name = 'image'
    filename, created_at = _image(
        images, out_path, col_name, ch_axis, row, mode, batched)

    value = kwargs
    value['timestamp'] = created_at.isoformat()
    value['images'] = {col_name: os.path.join(subdir, filename)}
    _chainerui_asset_observer.add(value)
    _chainerui_asset_observer.save(out_root, timeout)


def audio(audio, sample_rate, name=None, out=None, subdir='', timeout=5,
          **kwargs):
    """Summarize audio files to listen on a browser.

    An sampled array is converted as WAV audio file, saved to output directory,
    and reported to the ChainerUI server. The audio file is saved every called
    this function. The audio file will be listened on `assets` endpoint
    vertically. If need to aggregate audio files in row, use
    :func:`~chainerui.summary.reporter`.

    Example of how to set arguments::

       >>> from chainerui import summary
       >>> summary.set_out('/path/to/output')
       >>> rate = 44100
       >>>
       >>> summary.audio(sampled_array, rate, name='test')
       >>> # sampled_array can be listened on a browser.

    Add description about the audio file::

       >>> summary.image(
       >>>     sampled_array, rate, name='test', epoch=1, iteration=100)
       >>> # 'epoch' and 'iteration' column will be shown.

    Args:
        audio (:class:`numpy.ndarray` or :class:`cupy.ndarray` or \
            :class:`chainer.Variable`): sampled wave array.
        sample_rate (int): sampling rate.
        name (str): name of image. set as column name. when not setting,
            assigned ``'audio'``.
        out (str): directory path of output.
        subdir (str): sub-directory path of output.
        **kwargs (dict): key-value pair to show as description. regardless of
            empty or not, timestamp on created the image is added.
    """

    from chainerui.report.audio_report import check_available
    if not check_available():
        return
    from chainerui.report.audio_report import report as _audio

    out_root = _chainerui_asset_observer.get_outpath(out)
    out_path = os.path.join(out_root, subdir)
    if not os.path.isdir(out_path):
        os.makedirs(out_path)
    col_name = name
    if col_name is None:
        col_name = 'audio'
    filename, created_at = _audio(audio, sample_rate, out_path, col_name)

    value = kwargs
    value['timestamp'] = created_at.isoformat()
    value['audios'] = {col_name: os.path.join(subdir, filename)}
    _chainerui_asset_observer.add(value)
    _chainerui_asset_observer.save(out_root, timeout)


def text(text, name=None, out=None, timeout=5, **kwargs):
    """Summarize texts to show on a browser.

    Texts generated by training model is saved as asset and reported to the
    ChainerUI.

    Arguments:
        text (str): generated text.
        name (str): name of text. set as column name. when not setting,
            assigned ``'text'``.
        out (str): directory path of output.
        **kwargs (dict): key-value pair to show as description. regardless of
            empty or not, timestamp on created the text is added.
    """

    out_path = _chainerui_asset_observer.get_outpath(out)
    col_name = name
    if col_name is None:
        col_name = 'text'

    value = kwargs
    value['timestamp'] = datetime.datetime.now().isoformat()
    value['texts'] = {col_name: text}
    _chainerui_asset_observer.add(value)
    _chainerui_asset_observer.save(out_path, timeout)



import argparse
import json
import os
import shutil

from chainerui.utils.tempdir import tempdir


def convert_dict(conditions):
    if isinstance(conditions, argparse.Namespace):
        return vars(conditions)
    return conditions


def save_args(conditions, out_path):
    """A util function to save experiment condition for job table.

    Args:
        conditions (:class:`argparse.Namespace` or dict): Experiment conditions
            to show on a job table. Keys are show as table header and values
            are show at a job row.
        out_path (str): Output directory name to save conditions.

    """

    args = convert_dict(conditions)

    try:
        os.makedirs(out_path)
    except OSError:
        pass

    with tempdir(prefix='args', dir=out_path) as tempd:
        path = os.path.join(tempd, 'args.json')
        with open(path, 'w') as f:
            json.dump(args, f, indent=4)

        new_path = os.path.join(out_path, 'args')
        shutil.move(path, new_path)



from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String

from chainerui import database


class Snapshot(database.BASE):
    """Snapshot Model."""
    __tablename__ = 'snapshot'

    id = Column(Integer, primary_key=True)
    result_id = Column(Integer, ForeignKey('result.id'))
    name = Column(String(1024))
    iteration = Column(Integer)

    def __init__(self, name=None, iteration=None):
        self.name = name
        self.iteration = iteration

    def __repr__(self):
        return '<Snapshot id: %r />' % (self.id)

    @property
    def serialize(self):
        """serialize."""

        return {
            "id": self.id,
            "name": self.name,
            "iteration": self.iteration
        }



from chainerui.models.argument import Argument  # NOQA
from chainerui.models.asset import Asset  # NOQA
from chainerui.models.bindata import Bindata  # NOQA
from chainerui.models.command import Command  # NOQA
from chainerui.models.log import Log  # NOQA
from chainerui.models.project import Project  # NOQA
from chainerui.models.result import Result  # NOQA
from chainerui.models.snapshot import Snapshot  # NOQA



from chainerui.utils.is_jsonable import is_jsonable  # NOQA
from chainerui.utils.is_numberable import is_numberable  # NOQA
from chainerui.utils.log_report import LogReport  # NOQA
from chainerui.utils.save_args import save_args  # NOQA
from chainerui.utils.tempdir import tempdir  # NOQA



from math import isinf
from math import isnan

import msgpack
import numbers
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import LargeBinary

from chainerui import database


class Log(database.BASE):
    """Log Model."""
    __tablename__ = 'log'

    id = Column(Integer, primary_key=True)
    result_id = Column(Integer, ForeignKey('result.id'))
    data = Column(LargeBinary(2048))

    def __init__(self, data=None):
        bdata = msgpack.packb(data, use_bin_type=True)
        self.data = bdata

    def __repr__(self):
        return '<Log id: %r />' % (self.id)

    @property
    def serialize(self):
        """serialize."""

        log_dict = {}

        data = msgpack.unpackb(self.data, raw=False)
        for item in data.items():
            value_to_store = (
                None
                if not isinstance(item[1], numbers.Number)
                or isinf(item[1])
                or isnan(item[1])
                else item[1]
            )

            log_dict[item[0]] = value_to_store

        return {
            'id': self.id,
            'resultId': self.result_id,
            'logDict': log_dict
        }



import datetime
import os
import warnings

import chainer
from chainer import cuda

from chainerui.report.utils import get_hash
from chainerui.report.utils import get_unixtime


_write_wav = None


def _set_wav_writer():
    try:
        from scipy.io.wavfile import write

        def scipy_write_wav(out, data, rate):
            return write(out, rate, data)

        return scipy_write_wav
    except (ImportError, TypeError):
        return None


_write_wav = _set_wav_writer()
_available = _write_wav is not None


def check_available():
    if not _available:
        warnings.warn('Scipy is not installed on your environment, '
                      'so no audio file will be output at this time.'
                      'Please install Scipy to save WAV files.\n\n'
                      '  % pip install scipy\n')
    return _available


def report(audio, sample_rate, out, name):
    if isinstance(audio, chainer.Variable):
        audio = audio.data
    audio = cuda.to_cpu(audio)
    now = datetime.datetime.now()
    ts = get_unixtime(now)
    filename = '{}_{}.wav'.format(name, get_hash('{}'.format(ts)))
    filepath = os.path.join(out, filename)
    _write_wav(filepath, audio, sample_rate)

    return filename, now



"""init.

Revision ID: 213e2a3392f2
Revises:
Create Date: 2017-09-05 16:23:51.900998

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '213e2a3392f2'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('result',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('path_name',
                              sa.String(length=512), nullable=True),
                    sa.Column('name', sa.String(length=512), nullable=True),
                    sa.PrimaryKeyConstraint('id'),
                    sa.UniqueConstraint('path_name')
                    )
    op.create_table('argument',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('result_id', sa.Integer(), nullable=True),
                    sa.Column('data', sa.String(length=1024), nullable=True),
                    sa.ForeignKeyConstraint(['result_id'], ['result.id'], ),
                    sa.PrimaryKeyConstraint('id')
                    )
    op.create_table('command',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('result_id', sa.Integer(), nullable=True),
                    sa.Column('name', sa.String(length=1024), nullable=True),
                    sa.Column('body', sa.String(length=1024), nullable=True),
                    sa.ForeignKeyConstraint(['result_id'], ['result.id'], ),
                    sa.PrimaryKeyConstraint('id')
                    )
    op.create_table('log',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('result_id', sa.Integer(), nullable=True),
                    sa.Column('data', sa.String(length=1024), nullable=True),
                    sa.ForeignKeyConstraint(['result_id'], ['result.id'], ),
                    sa.PrimaryKeyConstraint('id')
                    )
    op.create_table('snapshot',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('result_id', sa.Integer(), nullable=True),
                    sa.Column('name', sa.String(length=1024), nullable=True),
                    sa.Column('iteration', sa.Integer(), nullable=True),
                    sa.ForeignKeyConstraint(['result_id'], ['result.id'], ),
                    sa.PrimaryKeyConstraint('id')
                    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('snapshot')
    op.drop_table('log')
    op.drop_table('command')
    op.drop_table('argument')
    op.drop_table('result')
    # ### end Alembic commands ###



from __future__ import absolute_import

import argparse
import logging
import os
import signal

from chainerui import _version
from chainerui.database import db
from chainerui.logging import logger
from chainerui.logging import set_loglevel
from chainerui.models.project import Project
from chainerui.server import create_app
from chainerui.utils import db_revision


def _check_db_revision():
    if not db_revision.check_current_db_revision():
        command = '\'chainerui db upgrade\''
        if db_revision.current_db_revision() is None:
            command = 'both \'chainerui db create\' and ' + command
        print('The current DB schema version is not supported.')
        print('Please run {} command before.'.format(command))
        return False
    return True


def _show_banner_debug(app, listener):
    # Such banner should be shown after running process of the server is
    # completed, but Flask has not supported to inject function after the
    # server started.
    from werkzeug.serving import is_running_from_reloader
    if is_running_from_reloader():
        # On debug mode, the banner is shown every reloaded.
        # run_simple set reloader type as 'stat' on default
        logger.info(' * Restarning with stat')
        # level warning is followed by werkzeug implementation
        logger.warning(' * Debugger is active!')
        from werkzeug.debug import get_pin_and_cookie_name
        pin, _ = get_pin_and_cookie_name(app)
        if pin is not None:
            logger.info(' * Debugger PIN: {:s}'.format(pin))
        return
    else:
        logger.info(' * Environment: {}'.format(app.config['ENV']))
        logger.info(' * Debug mode: on')
        logger.info(' * Running on http://{}/ (Press CTRL+C to quit)'.format(
            listener))


def server_handler(args):
    """server_handler."""
    if not db.setup(url=args.db, echo=args.db_echo):
        return
    if not _check_db_revision():
        return

    app = create_app()
    listener = '{:s}:{:d}'.format(args.host, args.port)
    if args.debug:
        logging.getLogger('werkzeug').disabled = True
        set_loglevel(logging.DEBUG)
        app.config['ENV'] = 'development'
        app.debug = True

        _show_banner_debug(app, listener)
        from werkzeug.serving import run_simple
        run_simple(
            args.host, args.port, app, use_reloader=True, use_debugger=True,
            threaded=True)
    else:
        app.config['ENV'] = 'production'
        import gevent
        from gevent.pywsgi import WSGIServer
        http_server = WSGIServer(listener, application=app, log=None)

        def stop_server():
            if http_server.started:
                http_server.stop()

        gevent.signal(signal.SIGTERM, stop_server)
        gevent.signal(signal.SIGINT, stop_server)
        logger.info(' * Environment: {}'.format(app.config['ENV']))
        logger.info(' * Running on http://{}/ (Press CTRL+C to quit)'.format(
            listener))

        try:
            http_server.serve_forever()
        except (KeyboardInterrupt, SystemExit):
            stop_server()


def db_handler(args):
    """db_handler."""

    if args.type == 'create':
        if args.db is None:
            db.init_db()
        return

    if not db.setup(url=args.db, echo=args.db_echo):
        return

    if args.type == 'status':
        current_rev = db_revision.current_db_revision()
        print('The current DB schema version:', current_rev)

    if args.type == 'upgrade':
        db.upgrade()

    if args.type == 'revision':
        db_revision.new_revision()

    if args.type == 'drop':
        if args.db is not None:
            db.downgrade()
        db.remove_db()


def project_create_handler(args):
    """project_create_handler."""
    if not db.setup(url=args.db, echo=args.db_echo):
        return
    if not _check_db_revision():
        return

    project_path = os.path.abspath(args.project_dir)
    project_name = args.project_name

    project = db.session.query(Project).\
        filter_by(path_name=project_path).first()

    if project is None:
        project = Project.create(project_path, project_name)
    else:
        print("Path '{}' has already registered.".format(project.path_name))


def create_parser():
    parser = argparse.ArgumentParser(description='chainerui command')
    parser.add_argument(
        '--version', '-v', action='version', version=_version.__version__)
    parser.add_argument(
        '--db', help='database resource address',
        default=os.getenv('CHAINERUI_DB_URL', default=None))
    parser.add_argument(
        '--db-echo', help='enable database enginge logging',
        action='store_true')
    subparsers = parser.add_subparsers()

    # chainerui server
    parser_server = subparsers.add_parser(
        'server', help='see `chainerui server -h`')
    parser_server.add_argument(
        '-H', '--host', required=False, help='host', default='localhost')
    parser_server.add_argument(
        '-p', '--port', required=False, type=int, help='port', default=5000)
    parser_server.add_argument(
        '-d', '--debug', action='store_true', help='debug')
    parser_server.set_defaults(
        handler=server_handler)

    # chainerui db
    parser_db = subparsers.add_parser(
        'db', help='see `chainerui db -h`')
    parser_db.add_argument(
        'type', choices=['create', 'drop', 'status', 'upgrade', 'revision'])
    parser_db.set_defaults(
        handler=db_handler)

    # chainerui project
    parser_project = subparsers.add_parser(
        'project', help='see `chainerui project -h`')
    parser_project_sub = parser_project.add_subparsers()

    # chainerui project create
    parser_project_create = parser_project_sub.add_parser(
        'create', help='see `chainerui project create -h`')
    parser_project_create.add_argument(
        '-d', '--project-dir', required=True, type=str, help='project-dir')
    parser_project_create.add_argument(
        '-n', '--project-name', type=str, help='project-name', default=None)
    parser_project_create.set_defaults(
        handler=project_create_handler)

    return parser


def main():
    """main."""
    parser = create_parser()
    args = parser.parse_args()

    if hasattr(args, 'handler'):
        args.handler(args)
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



"""empty message.

Revision ID: c78d4004717f
Revises: 6cc68bf684ac
Create Date: 2017-09-22 15:14:36.947242

"""


import datetime

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'c78d4004717f'
down_revision = '6cc68bf684ac'
branch_labels = None
depends_on = None

now = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")


def upgrade():
    with op.batch_alter_table('result') as batch_op:
        batch_op.add_column(
            sa.Column(
                'created_at',
                sa.DateTime(),
                nullable=False, server_default=now
            )
        )
        batch_op.add_column(
            sa.Column(
                'updated_at',
                sa.DateTime(),
                nullable=False, server_default=now
            )
        )

    with op.batch_alter_table('project') as batch_op:
        batch_op.add_column(
            sa.Column(
                'created_at',
                sa.DateTime(),
                nullable=False, server_default=now
            )
        )
        batch_op.add_column(
            sa.Column(
                'updated_at',
                sa.DateTime(),
                nullable=False, server_default=now
            )
        )


def downgrade():
    with op.batch_alter_table('result') as batch_op:
        batch_op.drop_column('created_at')
        batch_op.drop_column('updated_at')

    with op.batch_alter_table('project') as batch_op:
        batch_op.drop_column('created_at')
        batch_op.drop_column('updated_at')



import json
import urllib3


def urlopen(method, url, content_type='json', data=None, return_type='json'):
    """connect and get information via HTTP request

    A simplified wrapper for HTTP request, favor for JSON data format.
    When an error is occurred, error message is converted to text and return
    ``(None, text)`` tuple. This simplification is prefer especially to during
    training phase not to stop processes.

    Arguments:
        method (str): method type, GET/POST/PUT/DELETE
        url (str): target URL
        content_type (str): content type set on HTTP request header,
            ``'application/json'`` is set on default.
        data (obj): when ``content_type`` is ``'json'``, request data is dumped
            as JSON automatically.
        return_type (str): type name. returned value is converted JSON on
            default.

    Returns:
        tuple: ``(Response, '')`` when not error, ``(None, str)`` when error
            has occurred.
    """

    headers = {}
    body = data
    if content_type == 'json':
        headers['Content-Type'] = 'application/json'
        if data is not None:
            body = json.dumps(data)
    else:
        headers['Content-Type'] = content_type
    http = urllib3.PoolManager()
    try:
        res = http.request(method, url, body=body, headers=headers)
    except (urllib3.exceptions.NewConnectionError,
            urllib3.exceptions.MaxRetryError) as e:
        return None, 'cannot connect to the URL: {}, {}'.format(url, e)
    except urllib3.exceptions.HTTPError as e:
        return None, e.message

    str_data = res.data.decode('utf8')
    if return_type == 'json':
        try:
            obj = json.loads(str_data)
            if res.status != 200:
                return None, 'error: ' + obj.get('message', '')
        except ValueError:
            return None, 'error: ' + str_data
    else:
        obj = str_data

    return obj, ''



import datetime

from flask import jsonify
from flask import request
from flask.views import MethodView

from chainerui.database import db
from chainerui.models.log import Log
from chainerui.models.project import Project
from chainerui.models.result import Result


class LogAPI(MethodView):

    def post(self, project_id=None, result_id=None):
        project = db.session.query(Project).filter_by(id=project_id).first()
        if project is None:
            return jsonify({
                'project': None,
                'message': 'No interface defined for URL.'
            }), 404
        result = db.session.query(Result).filter_by(id=result_id).first()
        if result is None:
            return jsonify({
                'result': None,
                'message': 'No interface defined for URL.'
            }), 404

        data = request.get_json()
        log_json = data.get('log')
        modified_at = log_json.get('modifiedAt', None)
        if modified_at is not None:
            result.log_modified_at = datetime.datetime.fromtimestamp(
                modified_at)
        log_values = log_json.get('values', [])
        reset = log_json.get('reset', False)
        if reset:
            result.logs = []
        for value in log_values:
            result.logs.append(Log(value))

        db.session.commit()

        return jsonify({
            'logs': {
                'resultId': result.id,
                'insertedLogCount': len(log_values),
                'totalLogCount': len(result.logs)
            }
        })



from enum import Enum
import json
import os
import shutil

from chainerui.utils.command_item import CommandItem
from chainerui.utils.tempdir import tempdir


class JobStatus(Enum):
    INITIALIZED = 0
    RUNNING = 1
    STOPPED = 2
    NO_EXTENSION_ERROR = 3

    def __str__(self):
        return self.name.lower()


def _job_status_converter(o):
    if isinstance(o, JobStatus):
        return str(o)
    # pass other types to raise default exception


class CommandsState(object):

    _default_filename = '.chainerui_commands'

    @classmethod
    def run(cls, out_path):
        # NOTE: in future, optimizer information set in trainer will be addded
        #       to command sate, to customize "adjust hyperparameter" function
        out_path = out_path
        state = cls._load(out_path, initialize=True)
        state['job_status'] = JobStatus.RUNNING
        cls._dump(out_path, state)

    @classmethod
    def stop(cls, out_path):
        state = cls._load(out_path, initialize=True)
        if state['job_status'] != JobStatus.STOPPED:
            state['job_status'] = JobStatus.STOPPED
            cls._dump(out_path, state)

    @classmethod
    def job_status(cls, out_path):
        state = cls._load(out_path)
        if state is None:
            if os.path.isfile(CommandItem.commands_path(out_path)):
                # NOTE: this constraint is for back compatibility <= v0.1.1
                #       it is possible that set CommandsExtension but
                #       '.chainerui_commands' is not found. If '.commands'
                #       is found, judge as 'STOPPED'
                cls.stop(out_path)
                return cls.job_status(out_path)
            return JobStatus.NO_EXTENSION_ERROR
        return state['job_status']

    @classmethod
    def _load(cls, out_path, initialize=False):
        file_path = os.path.join(out_path, cls._default_filename)

        if os.path.isfile(file_path):
            with open(file_path, 'r') as f:
                state = json.load(f)
            state['job_status'] = JobStatus[state['job_status'].upper()]
        else:
            if initialize:
                state = {
                    'job_status': JobStatus.INITIALIZED
                }
                try:
                    os.makedirs(out_path)
                except OSError:
                    pass
            else:
                state = None
        return state

    @classmethod
    def _dump(cls, out_path, state):
        with tempdir(prefix=cls._default_filename, dir=out_path) as tempd:
            path = os.path.join(tempd, 'dump.json')
            with open(path, 'w') as f:
                json.dump(state, f, indent=4, default=_job_status_converter)

            shutil.move(path, os.path.join(out_path, cls._default_filename))



"""empty message

Revision ID: 9d002f65344b
Revises: c29d1b4b67fe
Create Date: 2018-10-08 18:26:28.588362

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '9d002f65344b'
down_revision = 'c29d1b4b67fe'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('project') as batch_op:
        batch_op.add_column(
            sa.Column('crawlable', sa.Boolean(), default=True))

    with op.batch_alter_table('result') as batch_op:
        batch_op.add_column(
            sa.Column('crawlable', sa.Boolean(), default=True))


def downgrade():
    with op.batch_alter_table('project') as batch_op:
        batch_op.drop_column('crawlable')

    with op.batch_alter_table('result') as batch_op:
        batch_op.drop_column('crawlable')



import json

from flask import jsonify
from flask import request
from flask.views import MethodView

from chainerui.database import db
from chainerui.models.argument import Argument
from chainerui.models.project import Project
from chainerui.models.result import Result


class ArgumentAPI(MethodView):

    def post(self, project_id=None, result_id=None):
        project = db.session.query(Project).filter_by(id=project_id).first()
        if project is None:
            return jsonify({
                'project': None,
                'message': 'No interface defined for URL.'
            }), 404
        result = db.session.query(Result).filter_by(id=result_id).first()
        if result is None:
            return jsonify({
                'result': None,
                'message': 'No interface defined for URL.'
            }), 404

        data = request.get_json()
        args = data.get('argument', {})
        result.args = Argument(json.dumps(args))

        db.session.commit()

        return jsonify({
            'argument': {'id': result.args.id, 'resultId': result.id}
        })



from __future__ import absolute_import
import datetime
import os

from flask import Flask
from flask import jsonify
from flask import render_template
from flask import request
from flask import send_from_directory
from flask import url_for
from sqlalchemy.exc import OperationalError

from chainerui.database import db
from chainerui.logging import logger


def create_app():
    """create_app."""

    app = Flask(__name__)
    app.logger.disabled = True
    for h in app.logger.handlers[:]:
        app.logger.removeHandler(h)
    app.config['JSONIFY_PRETTYPRINT_REGULAR'] = False

    def dated_url_for(endpoint, **values):
        """dated_url_for."""
        if endpoint == 'static':
            filename = values.get('filename', None)
            if filename:
                file_path = os.path.join(app.root_path, endpoint, filename)
                values['_'] = int(os.stat(file_path).st_mtime)
        return url_for(endpoint, **values)

    @app.context_processor
    def override_url_for():
        """override_url_for."""
        return dict(url_for=dated_url_for)

    @app.teardown_appcontext
    def shutdown_session(exception=None):
        db.session.remove()

    @app.route('/')
    @app.route('/projects/<int:project_id>')
    @app.route('/projects/<int:project_id>/results/<int:result_id>')
    @app.route('/projects/<int:project_id>/results/<int:result_id>/assets')
    def index(**kwargs):
        """render react app."""
        return render_template('index.html')

    @app.route('/favicon.ico')
    def favicon():
        return send_from_directory(
            os.path.join(app.root_path, 'static', 'dist'),
            'favicon.ico', mimetype='image/vnd.microsoft.icon')

    # error handling
    @app.errorhandler(OperationalError)
    def handle_invalid_usage(error):
        """handle errors caused by db query."""
        logger.error('caught exception from db: %s' % error.args)
        response = jsonify({
            'error': {
                'type': 'DBOperationalError',
                'message': 'Failed to send request to the database.'
            }
        })
        response.status_code = 400  # Bad Request

        return response

    @app.before_request
    def add_timestamp():
        request._comming_at = datetime.datetime.now()

    @app.after_request
    def output_log(response):
        now = datetime.datetime.now()
        log_msg = '%s - - [%s] "%s %s %s" %d' % (
            request.remote_addr, now.replace(microsecond=0),
            request.method, request.full_path,
            request.environ.get('SERVER_PROTOCOL'), response.status_code)
        if response.content_length is not None:
            log_msg += ' %d' % response.content_length
        if request._comming_at is not None:
            delta = (now - request._comming_at).total_seconds()
            log_msg += ' %.6f' % delta
        logger.info(log_msg)
        return response

    from chainerui.views.argument import ArgumentAPI
    from chainerui.views.log import LogAPI
    from chainerui.views.project import ProjectAPI
    from chainerui.views.result import ResultAPI
    from chainerui.views.result_asset import ResultAssetAPI
    from chainerui.views.result_command import ResultCommandAPI

    project_resource = ProjectAPI.as_view('project_resource')
    result_resource = ResultAPI.as_view('result_resource')
    log_resource = LogAPI.as_view('log_resource')
    arg_resource = ArgumentAPI.as_view('arg_resource')
    result_command_resource = ResultCommandAPI.as_view(
        'result_command_resource')
    result_assets_resource = ResultAssetAPI.as_view('result_assets_resource')

    # project API
    app.add_url_rule(
        '/api/v1/projects',
        defaults={'id': None}, view_func=project_resource, methods=['GET'])
    app.add_url_rule(
        '/api/v1/projects', view_func=project_resource, methods=['POST'])
    app.add_url_rule(
        '/api/v1/projects/<int:id>',
        view_func=project_resource, methods=['GET', 'PUT', 'DELETE'])

    # result API
    app.add_url_rule(
        '/api/v1/projects/<int:project_id>/results',
        defaults={'id': None}, view_func=result_resource, methods=['GET'])
    app.add_url_rule(
        '/api/v1/projects/<int:project_id>/results',
        view_func=result_resource, methods=['POST'])
    app.add_url_rule(
        '/api/v1/projects/<int:project_id>/results/<int:id>',
        view_func=result_resource, methods=['GET', 'PUT', 'DELETE'])

    # result log API
    app.add_url_rule(
        '/api/v1/projects/<int:project_id>/results/<int:result_id>/logs',
        view_func=log_resource, methods=['POST'])
    # result argument API
    app.add_url_rule(
        '/api/v1/projects/<int:project_id>/results/<int:result_id>/args',
        view_func=arg_resource, methods=['POST'])

    # result command API
    app.add_url_rule(
        '/api/v1/projects/<int:project_id>/results/<int:result_id>/commands',
        view_func=result_command_resource, methods=['POST'])

    # result image API
    app.add_url_rule(
        '/api/v1/projects/<int:project_id>/results/<int:result_id>/assets',
        view_func=result_assets_resource, methods=['GET'])
    app.add_url_rule(
        '/api/v1/projects/<int:project_id>/results/<int:result_id>/assets/<int:content_id>',  # NOQA
        view_func=result_assets_resource, methods=['GET'])

    return app



import hashlib
import time


def get_hash(key):
    return hashlib.md5(key.encode('utf-8')).hexdigest()[:12]


def get_unixtime(dt):
    # same as datetime.timestamp, but Python2.7 does not support the method.
    return time.mktime(dt.timetuple()) + dt.microsecond/1e6



from datetime import datetime
import json
import os
import shutil

from chainerui.utils.is_jsonable import is_jsonable
from chainerui.utils.tempdir import tempdir


class CommandItem(object):

    REQUEST_OPEN = 'OPEN'
    RESPONSE_SUCCESS = 'SUCCESS'
    RESPONSE_FAILURE = 'FAILURE'
    DEFAULT_FILE_NAME = 'commands'

    def __init__(self, **kwargs):
        self._name = kwargs.get('name', None)
        self._request = kwargs.get('request', None)
        self._response = kwargs.get('response', None)

    @property
    def name(self):
        return self._name

    @property
    def request(self):
        return self._request

    @property
    def response(self):
        return self._response

    @property
    def request_body(self):
        if self._request is None:
            return None
        return self._request.get('body', None)

    @property
    def response_body(self):
        if self._response is None:
            return None
        return self._response.get('body', None)

    def set_request(self, request_status, request_body, schedule):
        request = {
            'created_at': datetime.now().isoformat(),
            'status': request_status
        }

        if is_jsonable(request_body):
            request['body'] = request_body
        else:
            request['body'] = None

        if self.is_valid_schedule(schedule):
            request['schedule'] = schedule
        else:
            request['schedule'] = None

        self._request = request
        return request

    def set_response(self, trainer, response_status, response_body):
        response = {
            'executed_at': datetime.now().isoformat(),
            'epoch': trainer.updater.epoch,
            'iteration': trainer.updater.iteration,
            'elapsed_time': trainer.elapsed_time,
            'status': response_status
        }

        if not is_jsonable(response_body):
            response['body'] = None
        else:
            response['body'] = response_body

        self._response = response
        return response

    def should_execute(self, trainer):
        if self._response is not None:
            # already executed
            return False

        request = self._request
        if request is None:
            return False

        schedule = request.get('schedule', None)
        if schedule is not None:
            if schedule['key'] == 'epoch':
                if trainer.updater.epoch != schedule['value']:
                    return False
            elif schedule['key'] == 'iteration':
                if trainer.updater.iteration != schedule['value']:
                    return False
            else:
                # invalid schedule key
                return False

        return True

    @classmethod
    def is_valid_schedule(cls, schedule):
        if schedule is None:
            return True
        if schedule.get('key', None) in ['epoch', 'iteration'] \
                and isinstance(schedule.get('value', None), int):
            return True
        return False

    @classmethod
    def commands_path(cls, result_path, file_name=DEFAULT_FILE_NAME):
        return os.path.join(os.path.abspath(result_path), file_name)

    @classmethod
    def remove_commands_file(cls, result_path, file_name=DEFAULT_FILE_NAME):
        commands_path = cls.commands_path(result_path, file_name)
        if os.path.isfile(commands_path):
            os.remove(commands_path)

    @classmethod
    def load_commands(cls, result_path, file_name=DEFAULT_FILE_NAME):
        commands_path = cls.commands_path(result_path, file_name)
        commands = []

        if os.path.isfile(commands_path):
            with open(commands_path, 'r') as f:
                try:
                    commands = json.load(f)
                except ValueError:
                    pass

        return list(map(lambda cmd: cls(**cmd), commands))

    @classmethod
    def dump_commands(cls, commands, result_path, file_name=DEFAULT_FILE_NAME):
        commands_path = cls.commands_path(result_path, file_name)

        with tempdir(prefix=file_name, dir=result_path) as tempd:
            path = os.path.join(tempd, 'dump_commands.json')
            with open(path, 'w') as f:
                json.dump(list(map(lambda cmd: cmd.to_dict(), commands)),
                          f, indent=4)

            shutil.move(path, commands_path)

    def to_dict(self):
        return {
            'name': self._name,
            'request': self._request,
            'response': self._response
        }



"""empty message

Revision ID: e3db52a480f8
Revises: 00a907e5780f
Create Date: 2018-02-08 17:54:46.051323

"""
from alembic import op
import sqlalchemy as sa

import json

import msgpack


# revision identifiers, used by Alembic.
revision = 'e3db52a480f8'
down_revision = '00a907e5780f'
branch_labels = None
depends_on = None


def upgrade():
    conn = op.get_bind()

    temp_log_table = op.create_table(
        'temp_log',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('result_id', sa.Integer(), nullable=True),
        sa.Column('data', sa.LargeBinary(length=2048), nullable=True),
        sa.ForeignKeyConstraint(['result_id'], ['result.id'], ),
        sa.PrimaryKeyConstraint('id'))
    res = conn.execute('SELECT id, result_id, data FROM log')
    results = res.fetchall()
    if len(results) > 0:
        modified_logs = [{
            'id': r[0],
            'result_id': r[1],
            'data': msgpack.packb(json.loads(r[2]), use_bin_type=True)}
            for r in results]
        op.bulk_insert(temp_log_table, modified_logs)
    op.drop_table('log')
    op.rename_table('temp_log', 'log')


def downgrade():
    conn = op.get_bind()

    temp_log_table = op.create_table(
        'temp_log',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('result_id', sa.Integer(), nullable=True),
        sa.Column('data', sa.String(length=1024), nullable=True),
        sa.ForeignKeyConstraint(['result_id'], ['result.id'], ),
        sa.PrimaryKeyConstraint('id'))
    res = conn.execute('SELECT id, result_id, data FROM log')
    results = res.fetchall()
    if len(results) > 0:
        modified_logs = [{
            'id': r[0],
            'result_id': r[1],
            'data': json.dumps(msgpack.unpackb(r[2], raw=False))}
            for r in results]
        op.bulk_insert(temp_log_table, modified_logs)
    op.drop_table('log')
    op.rename_table('temp_log', 'log')



import datetime
import json
import os

from chainerui.database import db
from chainerui.logging import logger
from chainerui.models.argument import Argument
from chainerui.models.command import Command
from chainerui.models.log import Log
from chainerui.models.snapshot import Snapshot
from chainerui.utils import is_numberable


def load_result_json(result_path, json_file_name):
    """load_result_json."""
    json_path = os.path.join(result_path, json_file_name)

    _list = []
    if os.path.isfile(json_path):
        with open(json_path) as json_data:
            try:
                _list = json.load(json_data)
            except ValueError as err:
                logger.error(
                    'Failed to load json: {}, {}'.format(json_path, err))

    return _list


def crawl_result_path(result_path, include_log):
    """crawl_result_path."""
    result = {
        'logs': [],
        'args': [],
        'commands': [],
        'snapshots': []
    }

    if os.path.isdir(result_path):
        if include_log:
            result['logs'] = load_result_json(result_path, 'log')
        result['args'] = load_result_json(result_path, 'args')
        result['commands'] = load_result_json(result_path, 'commands')

        snapshots = [
            x for x in os.listdir(result_path) if x.count('snapshot_iter_')
        ]
        snapshots.sort()
        result['snapshots'] = snapshots

    return result


def _check_log_updated(result):
    log_json_path = os.path.join(result.path_name, 'log')
    if not os.path.isfile(log_json_path):
        # log file is removed, so don't have to update
        return False

    current_modified_at = result.log_modified_at
    modified_at = datetime.datetime.fromtimestamp(
        os.path.getmtime(log_json_path))
    if current_modified_at is None or current_modified_at != modified_at:
        result.log_modified_at = modified_at
        return True

    return False


def _update_to_default_name(result):
    conf_path = os.path.join(result.path_name, '.chainerui_conf')
    if not os.path.isfile(conf_path):
        return
    with open(conf_path) as f:
        try:
            chainerui_conf = json.load(f)
        except ValueError as err:
            logger.error(
                'Failed to load json: {}, {}'.format(conf_path, err))
            return
    default_name = chainerui_conf.get('default_result_name', None)
    if default_name is None:
        return
    result.name = default_name


def crawl_result(result, force=False, commit=True):
    """crawl_results."""
    if not result.crawlable:
        return result
    now = datetime.datetime.now()

    if (not force) and (now - result.updated_at).total_seconds() < 4:
        return result

    # if log file is not updated, not necessary to get log contents
    is_updated = _check_log_updated(result)
    crawled_result = crawl_result_path(result.path_name, is_updated)

    if is_updated:
        current_log_idx = len(result.logs)
        if len(crawled_result['logs']) < current_log_idx:
            current_log_idx = 0
            result.logs = []
            result.args = None
        for log in crawled_result['logs'][current_log_idx:]:
            result.logs.append(Log(log))

    if result.args is None:
        result.args = Argument(json.dumps(crawled_result['args']))

    if result.name is None:
        _update_to_default_name(result)

    # commands list includes new commands and already registered commands.
    # registered commands can be get response, so need to update
    current_cmd_idx = len(result.commands)
    if len(crawled_result['commands']) < current_cmd_idx:
        current_cmd_idx = 0
        result.commands = []
        result.snapshots = []
    for cmd in crawled_result['commands'][current_cmd_idx:]:
        result.commands.append(Command(**cmd))
    for i, cmd in enumerate(crawled_result['commands'][:current_cmd_idx]):
        result.commands[i].update(cmd.get('response', None))

    # snapshots file list are sorted but not natural order, for example,
    # 'iter_900' set latter than 'iter_1000', so need to check the file
    # is registered or not.
    registered_snapshot_keys = [ss.iteration for ss in result.snapshots]
    for i, snapshot in enumerate(crawled_result['snapshots']):
        number_str = snapshot.split('snapshot_iter_')[1]
        if not is_numberable(number_str):
            continue
        number = int(number_str)
        if number in registered_snapshot_keys:
            continue
        result.snapshots.append(Snapshot(snapshot, number))

    result.updated_at = datetime.datetime.now()
    if commit:
        db.session.commit()

    return result



"""empty message

Revision ID: 00a907e5780f
Revises: c78d4004717f
Create Date: 2018-01-31 12:19:39.490933

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '00a907e5780f'
down_revision = 'c78d4004717f'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('result') as batch_op:
        batch_op.add_column(
            sa.Column(
                'log_modified_at',
                sa.DateTime(),
                nullable=True
            )
        )


def downgrade():
    with op.batch_alter_table('result') as batch_op:
        batch_op.drop_column('log_modified_at')



"""empty message

Revision ID: 6cd7f193a3de
Revises: e3db52a480f8
Create Date: 2018-06-15 17:22:02.936060

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '6cd7f193a3de'
down_revision = 'e3db52a480f8'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'asset',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('result_id', sa.Integer(), nullable=False),
        sa.Column('summary', sa.String(length=1024), nullable=True),
        sa.Column('file_modified_at', sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(['result_id'], ['result.id'], ),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'bindata',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('asset_id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(length=1024), nullable=False),
        sa.Column('tag', sa.String(length=1024), nullable=True),
        sa.Column('note', sa.String(length=1024), nullable=True),
        # < 10MB
        sa.Column('content', sa.LargeBinary(length=1e7), nullable=False),
        sa.ForeignKeyConstraint(['asset_id'], ['asset.id'], ),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('bindata')
    op.drop_table('asset')



import os
import shutil

from chainer.serializers import npz
from chainer.training import extension
from chainer.training import trigger as trigger_module
from chainer.training.triggers import IntervalTrigger
import six

from chainerui.utils.command_item import CommandItem
from chainerui.utils.commands_state import CommandsState
from chainerui.utils.tempdir import tempdir


def take_snapshot(trainer, body):
    filename = 'snapshot_iter_{.updater.iteration}'.format(trainer)
    out_path = trainer.out
    # same with SimpleWriter, supported from Chainer v6
    with tempdir(prefix='snapshot', dir=out_path) as tempd:
        path = os.path.join(tempd, filename)
        npz.save_npz(path, trainer)
        shutil.move(path, os.path.join(out_path, filename))


def adjust_hyperparams(trainer, body):
    optimizer = trainer.updater.get_optimizer('main')
    optimizer_name = optimizer.__class__.__name__
    if optimizer_name != body.get('optimizer', None):
        # invalid optimizer was specified
        return None

    hyperparam = getattr(optimizer, 'hyperparam', None)
    if hyperparam is None:
        return None

    request_hyperparam = body.get('hyperparam', {})
    hyperparam_dict = hyperparam.get_dict()
    for key, value in six.iteritems(request_hyperparam):
        if (key not in hyperparam_dict) or (value is None):
            continue  # pragma: no cover
        setattr(hyperparam, key, value)
    return {
        'optimizer': optimizer_name,
        'hyperparam': hyperparam.get_dict()
    }

# NOTE: Chainer has a plan to add that trigger can detect training
#       length (PR#4079). After merge it, the below two trigger class
#       can be merge to one trigger class.


class _CommandIntervalTrigger(IntervalTrigger):

    def __init__(self, trigger):
        super(_CommandIntervalTrigger, self).__setattr__(
            '_trigger', trigger)
        super(_CommandIntervalTrigger, self).__setattr__(
            '_loop_stop', False)

    def __call__(self, trainer):
        if self._trigger(trainer):
            return True
        return self._loop_stop

    def stop(self):
        super(_CommandIntervalTrigger, self).__setattr__(
            '_loop_stop', True)

    def __getattr__(self, attr_name):
        return getattr(self._trigger, attr_name)

    def __setattr__(self, attr_name, value):
        setattr(self._trigger, attr_name, value)


class _CommandTrigger(object):

    def __init__(self, trigger):
        super(_CommandTrigger, self).__setattr__(
            '_trigger', trigger)
        super(_CommandTrigger, self).__setattr__(
            '_loop_stop', False)

    def __call__(self, trainer):
        if self._trigger(trainer):
            return True
        return self._loop_stop

    def stop(self):
        super(_CommandTrigger, self).__setattr__(
            '_loop_stop', True)

    def __getattr__(self, attr_name):
        return getattr(self._trigger, attr_name)

    def __setattr__(self, attr_name, value):
        setattr(self._trigger, attr_name, value)


def _stop_training(trainer, body):
    assert isinstance(trainer.stop_trigger, _CommandTrigger) or \
        isinstance(trainer.stop_trigger, _CommandIntervalTrigger)
    trainer.stop_trigger.stop()
    return None


class CommandsExtension(extension.Extension):

    """Trainer extension to enable command operation by output file

    This extension monitors a file for command created on `trainer.out` path,
    and execute each command when append the file.

    """

    priority = extension.PRIORITY_READER
    default_receivers = {
        'take_snapshot': take_snapshot,
        'adjust_hyperparams': adjust_hyperparams,
        'stop': _stop_training,
    }

    def __init__(self, trigger=(1, 'iteration'), receivers={},
                 file_name='commands'):
        self._trigger = trigger_module.get_trigger(trigger)
        self._file_name = file_name

        self._receivers = self.default_receivers.copy()
        self._receivers.update(receivers)

        self._out = ''

    def initialize(self, trainer):
        self._out = trainer.out
        CommandItem.remove_commands_file(trainer.out)
        CommandsState.run(trainer.out)
        if isinstance(trainer.stop_trigger, IntervalTrigger):
            trainer.stop_trigger = _CommandIntervalTrigger(
                trainer.stop_trigger)
        else:
            trainer.stop_trigger = _CommandTrigger(trainer.stop_trigger)

    def __call__(self, trainer):
        if not self._trigger(trainer):
            return

        commands = CommandItem.load_commands(trainer.out)

        is_updated = False
        for command in commands:
            if not command.should_execute(trainer):
                continue

            body, status = self._execute_command(
                trainer, command.name, command.request)
            command.set_response(trainer, status, body)
            is_updated = True

        if is_updated:
            CommandItem.dump_commands(commands, trainer.out)

    def finalize(self):
        if self._out != '':
            CommandsState.stop(self._out)

    def add_receiver(self, command_name, function):
        if command_name is None:
            raise ValueError('command name is not given')
        if not callable(function):
            raise ValueError('receiver is not callable')
        self._receivers[command_name] = function

    def _execute_command(self, trainer, command_name, request):
        receiver = self._receivers.get(command_name, None)
        if receiver is None:
            message = '%s command is not available or supported' % command_name
            response_body = {'message': message}
            response_status = CommandItem.RESPONSE_FAILURE
        else:
            try:
                response_body = receiver(trainer, request.get('body', None))
                response_status = CommandItem.RESPONSE_SUCCESS
            except Exception as e:
                print('caught an exception from receiver:', e.args)
                response_body = None
                response_status = CommandItem.RESPONSE_FAILURE

        return response_body, response_status



from collections import OrderedDict
import datetime
import json
import os

from chainerui.database import db
from chainerui.models.asset import Asset
from chainerui.models.bindata import Bindata
from chainerui import summary


def collect_assets(result, force=False):
    """collect assets from meta file

    Collecting assets only when the metafile is updated. If number of assets
    are decreased, assets are reset and re-collect the assets.
    """
    path_name = result.path_name
    info_path = os.path.join(path_name, summary.CHAINERUI_ASSETS_METAFILE_NAME)
    if not os.path.isfile(info_path):
        return
    start_idx = len(result.assets)
    file_modified_at = datetime.datetime.fromtimestamp(os.path.getmtime(
        info_path))
    if start_idx > 0:
        if result.assets[-1].file_modified_at == file_modified_at:
            return

    with open(info_path, 'r') as f:
        info_list = json.load(f, object_pairs_hook=OrderedDict)

    if len(info_list) < start_idx:
        start_idx = 0
        result.assets = []

    for base_info in info_list[start_idx:]:
        asset_path = base_info.pop('images', {})
        asset_path.update(base_info.pop('audios', {}))
        base_info.update(base_info.pop('texts', {}))
        asset = Asset.create(
            result_id=result.id, summary=base_info,
            file_modified_at=file_modified_at)
        for key, path in asset_path.items():
            with open(os.path.join(path_name, path), 'rb') as f:
                data = f.read()
            content = Bindata(
                asset_id=asset.id, name=path, tag=key, content=data)
            asset.content_list.append(content)
        result.assets.append(asset)

    db.session.commit()



import io

from flask import jsonify
from flask import send_file
from flask.views import MethodView

from chainerui.database import db
from chainerui.models.bindata import Bindata
from chainerui.models.project import Project
from chainerui.models.result import Result
from chainerui.tasks.collect_assets import collect_assets


class ResultAssetAPI(MethodView):

    def get(self, result_id=None, project_id=None, content_id=None):
        # project is not necessary to collect assets,
        # but check parent project exists or not
        project = db.session.query(Project).\
            filter_by(id=project_id).\
            first()
        if project is None:
            return jsonify({
                'project': None,
                'message': 'No interface defined for URL.'
            }), 404

        result = db.session.query(Result).\
            filter_by(id=result_id).\
            filter_by(is_unregistered=False).\
            first()
        if result is None:
            return jsonify({
                'result': None,
                'message': 'No interface defined for URL.'
            }), 404

        if content_id is None:
            collect_assets(result)
            assets_response = [asset.serialize for asset in result.assets]
            for asset in assets_response:
                for content in asset['contents']:
                    content['uri'] = self._make_content_uri(
                        project_id, result_id, content['id'])
            return jsonify({'assets': assets_response})

        # sent content binary directly
        bindata = db.session.query(Bindata).\
            filter_by(id=content_id).\
            first()
        if bindata is None:
            return jsonify({
                'asset': None,
                'message': 'No interface defined for URL.'
            }), 404
        return send_file(
            io.BytesIO(bindata.content),
            mimetype=bindata.mimetype(),
            as_attachment=True,
            attachment_filename=bindata.name)

    def _make_content_uri(self, project_id, result_id, content_id):
        return '/api/v1/projects/%d/results/%d/assets/%d' % (
            project_id, result_id, content_id)



from flask import jsonify
from flask import request
from flask.views import MethodView

from chainerui.database import db
from chainerui.models.result import Result
from chainerui.tasks import crawl_result
from chainerui.utils.command_item import CommandItem
from chainerui.utils.commands_state import CommandsState
from chainerui.utils.commands_state import JobStatus


class ResultCommandAPI(MethodView):
    """ResultCommandAPI."""

    def post(self, result_id, project_id):
        """POST /api/v1/results/<int:id>/commands."""

        result = db.session.query(Result).filter_by(id=result_id).first()

        if result is None:
            return jsonify({
                'result': None,
                'message': 'No interface defined for URL.'
            }), 404

        job_status = CommandsState.job_status(result.path_name)
        if job_status != JobStatus.RUNNING:
            if job_status == JobStatus.NO_EXTENSION_ERROR:
                return jsonify({
                    'message': '\'CommandsExtension\' is not set or disabled.'
                }), 400
            elif job_status == JobStatus.INITIALIZED:
                return jsonify({
                    'message': 'The target training job has not run, yet'
                }), 400
            elif job_status == JobStatus.STOPPED:
                return jsonify({
                    'message': 'The target training job has already stopped'
                }), 400
            else:
                return jsonify({
                    'message': 'Cannot get the target training job status'
                }), 400

        request_json = request.get_json()
        if request_json is None:
            return jsonify({
                'message': 'Empty request.'
            }), 400

        command_name = request_json.get('name', None)
        if command_name is None:
            return jsonify({
                'message': 'Name is required.'
            }), 400

        schedule = request_json.get('schedule', None)
        if not CommandItem.is_valid_schedule(schedule):
            return jsonify({
                'message': 'Schedule is invalid.'
            }), 400

        command = CommandItem(
            name=command_name,
        )

        command.set_request(
            CommandItem.REQUEST_OPEN,
            request_json.get('body', None),
            request_json.get('schedule', None)
        )

        commands = CommandItem.load_commands(result.path_name)
        commands.append(command)

        CommandItem.dump_commands(commands, result.path_name)

        new_result = crawl_result(result, force=True)
        new_result_dict = new_result.serialize

        return jsonify({'commands': new_result_dict['commands']})



import datetime

from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy.orm import relationship
from sqlalchemy.sql.expression import asc
from sqlalchemy import String

from chainerui import database
from chainerui.database import db
from chainerui.models.log import Log
from chainerui.tasks.crawl_result import crawl_result


class Result(database.BASE):
    """Result Model."""
    __tablename__ = 'result'

    id = Column(Integer, primary_key=True)
    project_id = Column(Integer, ForeignKey('project.id'))
    path_name = Column(String(512), unique=True)
    name = Column(String(512))
    is_unregistered = Column(Boolean(), default=False)
    logs = relationship(
        'Log', cascade='all, delete-orphan', order_by=lambda: asc(Log.id))
    args = relationship(
        'Argument', uselist=False, cascade='all, delete-orphan')
    commands = relationship('Command', cascade='all, delete-orphan')
    snapshots = relationship('Snapshot', cascade='all, delete-orphan')
    assets = relationship('Asset', cascade='all, delete-orphan')
    log_modified_at = Column(DateTime, default=None)
    crawlable = Column(Boolean(), default=True)
    created_at = Column(DateTime, default=datetime.datetime.now())
    updated_at = Column(DateTime, default=datetime.datetime.now())

    def __init__(self, path_name=None, name=None, project_id=None,
                 log_modified_at=None, crawlable=True):
        self.path_name = path_name
        self.name = name
        self.project_id = project_id
        self.log_modified_at = log_modified_at
        self.crawlable = crawlable

    def __repr__(self):
        return '<Result id: %r, path_name: %r />' % (self.id, self.path_name)

    @classmethod
    def create(cls, path_name=None, name=None, project_id=None,
               log_modified_at=None, crawlable=True):
        """Initialize an instance and save it to db."""
        result = cls(path_name, name, project_id, log_modified_at, crawlable)

        db.session.add(result)
        db.session.commit()

        crawl_result(result, True)

        return result

    def sampled_logs(self, logs_limit=-1):
        """Return up to `logs_limit` logs.

        If `logs_limit` is -1, this function will return all logs that belong
        to the result.
        """
        logs_count = len(self.logs)
        if logs_limit == -1 or logs_count <= logs_limit:
            return self.logs
        elif logs_limit == 0:
            return []
        elif logs_limit == 1:
            return [self.logs[-1]]
        else:
            def get_sampled_log(idx):
                # always include the first and last element of `self.logs`
                return self.logs[idx * (logs_count - 1) // (logs_limit - 1)]
            return [get_sampled_log(i) for i in range(logs_limit)]

    def serialize_with_sampled_logs(self, logs_limit=-1):
        """serialize a result with up to `logs_limit` logs.

        If `logs_limit` is -1, this function will return a result with all its
        logs.
        """

        return {
            'id': self.id,
            'pathName': self.path_name,
            'name': self.name,
            'isUnregistered': self.is_unregistered,
            'logs': [log.serialize for log in self.sampled_logs(logs_limit)],
            'args': self.args.serialize if self.args is not None else [],
            'commands': [cmd.serialize for cmd in self.commands],
            'snapshots': [cmd.serialize for cmd in self.snapshots],
            'logModifiedAt': self.log_modified_at.isoformat()
        }

    @property
    def serialize(self):
        """serialize."""

        return self.serialize_with_sampled_logs(-1)



import contextlib
import shutil
import tempfile


@contextlib.contextmanager
def tempdir(**kwargs):
    # A context manager that defines a lifetime of a temporary directory.
    ignore_errors = kwargs.pop('ignore_errors', False)

    temp_dir = tempfile.mkdtemp(**kwargs)
    try:
        yield temp_dir
    finally:
        shutil.rmtree(temp_dir, ignore_errors=ignore_errors)



import datetime
import os
import warnings

import chainer
from chainer import cuda
import numpy

from chainerui.report.utils import get_hash
from chainerui.report.utils import get_unixtime


def _setup_image_module():
    try:
        from PIL import Image  # NOQA
        return Image

    except (ImportError, TypeError):
        return None


_Image = _setup_image_module()
_available = _Image is not None


def check_available():
    if not _available:
        warnings.warn('Pillow is not installed on your environment, '
                      'so no image will be output at this time.'
                      'Please install Pillow to save images.\n\n'
                      '  $ pip install Pillow\n')
    return _available


def report(images, out, name, ch_axis=1, row=0, mode=None, batched=True):
    if isinstance(images, chainer.Variable):
        images = images.data
    images = cuda.to_cpu(images)
    if batched:
        stuck_image = _get_stuck_batched_image(images, ch_axis, row)
    else:
        stuck_image = _get_stuck_image(images, ch_axis)

    now = datetime.datetime.now()
    ts = get_unixtime(now)
    filename = '{}_{}.png'.format(name, get_hash('{}'.format(ts)))
    filepath = os.path.join(out, filename)
    _save_image(_normalize_8bit(stuck_image), filepath, mode=mode)

    return filename, now


def _get_stuck_batched_image(images, ch_axis, row):
    ndim = images.ndim
    if not (ndim == 3 or ndim == 4):
        raise ValueError(
            'Number of array dimension {:d} must be 3 or 4'.format(ndim))

    if ndim == 4:
        images = _move_ch_to_last(images, ch_axis)
        B, H, W, C = images.shape
        if row == 0:
            row = B
        col = B // row
        images = images.reshape(row, col, H, W, C)
        images = images.transpose(0, 2, 1, 3, 4)
        return images.reshape(row * H, col * W, C)

    # ndim == 3
    B, H, W = images.shape
    if row == 0:
        row = B
    col = B // row
    images = images.reshape(row, col, H, W)
    images = images.transpose(0, 2, 1, 3)
    return images.reshape(row * H, col * W)


def _get_stuck_image(images, ch_axis):
    ndim = images.ndim
    if not (ndim == 2 or ndim == 3):
        raise ValueError(
            'Number of array dimension {:d} must be 2 or 3'.format(ndim))
    if ndim == 2:
        return images

    # ndim == 3
    return _move_ch_to_last(images, ch_axis)


def _move_ch_to_last(x, axis):
    if axis == -1:
        return x
    rolled_ax = numpy.append(numpy.delete(numpy.arange(x.ndim), axis), axis)
    return x.transpose(rolled_ax)


def _normalize_8bit(array):
    if array.dtype == numpy.uint8:
        return array
    return numpy.asarray(numpy.clip(array*255, 0.0, 255.0), dtype=numpy.uint8)


def _save_image(img, name, ext='PNG', mode=None):
    if mode is None:
        _Image.fromarray(img).save(name, format=ext)
    elif mode.lower() == 'hsv':
        _Image.fromarray(img, mode='HSV').convert('RGB').save(name, format=ext)



from alembic.command import revision
from alembic.migration import MigrationContext
from alembic.script import ScriptDirectory

from chainerui.database import db


def current_db_revision():
    ctx = MigrationContext.configure(db.engine.connect())
    return ctx.get_current_revision()


def check_current_db_revision():
    current_rev = current_db_revision()
    if current_rev is None:
        return False

    config = db.alembic_config
    script = ScriptDirectory.from_config(config)
    module_rev = script.get_current_head()
    if current_rev != module_rev:
        return False

    return True


def new_revision():
    config = db.alembic_config
    revision(config)



import datetime

from flask import jsonify
from flask import request
from flask.views import MethodView

from chainerui.database import db
from chainerui.models.project import Project
from chainerui.models.result import Result
from chainerui.tasks import collect_results
from chainerui.tasks import crawl_result


class ResultAPI(MethodView):
    """ResultAPI."""

    def get(self, id=None, project_id=None):
        """get."""
        logs_limit = request.args.get('logs_limit', default=-1, type=int)
        is_unregistered = request.args.get(
            'is_unregistered', default=False, type=bool)

        project = db.session.query(Project).filter_by(
            id=project_id).first()
        if project is None:
            return jsonify({
                'project': None,
                'message': 'No interface defined for URL.'
            }), 404

        if id is None:
            path = request.args.get('path_name', default=None)
            if path is not None:
                result = db.session.query(Result).filter_by(
                    path_name=path).first()
                if result is None:
                    return jsonify({
                        'result': None,
                        'message': 'Result path \'%s\' is not found' % path
                    }), 400
                return jsonify({'result': result.serialize})

            collect_results(project)

            results = db.session.query(Result).\
                filter_by(project_id=project_id).\
                filter_by(is_unregistered=is_unregistered).\
                all()

            # NOTE: To improve performance, aggregate commit phase. By set
            # `commit=False`, implicit transaction is not closed, UPDATE query
            # is not committed. Consequently a process of serializing does not
            # have to call SELECT query again.
            for result in results:
                crawl_result(result, commit=False)
            db.session.commit()
            rs = [r.serialize_with_sampled_logs(logs_limit) for r in results]

            return jsonify({'results': rs})

        else:

            result = db.session.query(Result).\
                filter_by(id=id).\
                filter_by(is_unregistered=False).\
                first()

            if result is None:
                return jsonify({
                    'result': None,
                    'message': 'No interface defined for URL.'
                }), 404

            result = crawl_result(result)

            return jsonify({
                'result': result.serialize_with_sampled_logs(logs_limit)
            })

    def post(self, project_id=None):
        project = db.session.query(Project).filter_by(id=project_id).first()
        if project is None:
            return jsonify({
                'project': None,
                'message': 'No interface defined for URL.'
            }), 404

        data = request.get_json()
        result_json = data.get('result')
        path = result_json.get('pathName', '')
        if path == '':
            return jsonify({
                'result': None,
                'message': 'Path of the result is not set.'
            }), 400
        result = db.session.query(Result).filter_by(path_name=path).first()
        if result is not None:
            return jsonify({
                'result': None,
                'message': 'Result path \'%s\' already registered.' % path
            }), 400

        name = result_json.get('name', None)
        crawlable = result_json.get('crawlable', True)
        log_modified_at = result_json.get('logModifiedAt', None)
        if log_modified_at is not None:
            log_modified_at = datetime.datetime.fromtimestamp(log_modified_at)
        result = Result.create(
            path_name=path, name=name, project_id=project_id,
            log_modified_at=log_modified_at, crawlable=crawlable)
        # don't return all data to reduce data size
        return jsonify({
            'result': {'id': result.id}
        })

    def put(self, id, project_id=None):
        """put."""
        result = db.session.query(Result).filter_by(id=id).first()
        if result is None:
            response = jsonify({
                'result': None, 'message': 'No interface defined for URL.'
            })
            return response, 404

        request_json = request.get_json()
        request_result = request_json.get('result')

        name = request_result.get('name', None)
        if name is not None:
            result.name = name

        is_unregistered = request_result.get('isUnregistered', None)
        if is_unregistered is not None:
            result.is_unregistered = is_unregistered

        db.session.add(result)
        db.session.commit()

        return jsonify({'result': result.serialize})

    def delete(self, id, project_id=None):
        """delete."""
        result = db.session.query(Result).filter_by(id=id).first()
        if result is None:
            response = jsonify({
                'result': None, 'message': 'No interface defined for URL.'
            })
            return response, 404

        db.session.delete(result)
        db.session.commit()

        return jsonify({'result': result.serialize})



"""empty message.

Revision ID: 7fcaf5b6e9e8
Revises: 213e2a3392f2
Create Date: 2017-09-13 17:58:59.577263

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '7fcaf5b6e9e8'
down_revision = '213e2a3392f2'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('command') as batch_op:
        batch_op.add_column(sa.Column('request', sa.String(1024)))
        batch_op.add_column(sa.Column('response', sa.String(1024)))
        batch_op.drop_column('body')


def downgrade():
    with op.batch_alter_table('command') as batch_op:
        batch_op.drop_column('request')
        batch_op.drop_column('response')
        batch_op.add_column(sa.Column('body', sa.String(1024), nullable=True))



from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import LargeBinary
from sqlalchemy import String

from chainerui import database


class Bindata(database.BASE):
    __tablename__ = 'bindata'

    id = Column(Integer, primary_key=True)
    asset_id = Column(Integer, ForeignKey('asset.id'))
    name = Column(String(512))
    tag = Column(String(512))
    note = Column(String(512))
    content = Column(LargeBinary(1e7))

    def __init__(self, asset_id=None, name=None, tag=None, note=None,
                 content=None):
        self.asset_id = asset_id
        self.name = name
        self.tag = tag
        self.note = note
        self.content = content

    def __repr__(self):
        return '<Bindata id: %r />' % (self.id)

    def mimetype(self):
        ext = self.name.split('.')[-1].lower()
        if ext == 'png':
            return 'image/png'
        elif ext in ['jpg', 'jpeg']:
            return 'image/jpeg'
        elif ext == 'wav':
            return 'audio/wav'
        else:
            raise ValueError('"%s" is not support' % ext)

    @property
    def serialize(self):
        # omit content to reduce transport size
        return {
            'id': self.id,
            'asset_id': self.asset_id,
            'name': self.name,
            'tag': self.tag,
            'note': self.note
        }



import datetime

from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import Integer
from sqlalchemy.orm import relationship
from sqlalchemy import String

from chainerui import database
from chainerui.database import db
from chainerui.tasks.collect_results import collect_results


class Project(database.BASE):
    """Project Model."""
    __tablename__ = 'project'

    id = Column(Integer, primary_key=True)
    path_name = Column(String(512), unique=True)
    name = Column(String(512))
    results = relationship('Result', cascade='all, delete-orphan')
    crawlable = Column(Boolean(), default=True)
    created_at = Column(DateTime, default=datetime.datetime.now())
    updated_at = Column(DateTime, default=datetime.datetime.now())

    def __init__(self, path_name=None, name=None, crawlable=True):
        self.path_name = path_name
        self.name = name
        self.crawlable = crawlable

    def __repr__(self):
        return '<Project id: %r, path_name: %r />' % (self.id, self.path_name)

    @classmethod
    def create(cls, path_name=None, name=None, crawlable=True):
        """initialize an instance and save it to db."""

        project = cls(path_name, name, crawlable)

        db.session.add(project)
        db.session.commit()

        return collect_results(project, force=True)

    @property
    def serialize(self):
        """serialize."""
        return {
            'id': self.id,
            'pathName': self.path_name,
            'name': self.name
        }



import json

from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String

from chainerui import database


class Argument(database.BASE):
    """Argument Model."""
    __tablename__ = 'argument'

    id = Column(Integer, primary_key=True)
    result_id = Column(Integer, ForeignKey('result.id'))
    data = Column(String(1024))

    def __init__(self, data=None):
        self.data = data

    def __repr__(self):
        return '<Argument id: %r />' % (self.id)

    @property
    def serialize(self):
        """serialize."""

        arguments = []

        if isinstance(json.loads(self.data), dict):
            for k, v in json.loads(self.data).items():
                arguments.append({
                    'resultId': self.result_id,
                    'key': k,
                    'value': str(v) if v is not None else None,
                })

        return arguments



"""add_is_unregistered_to_result.

Revision ID: 4e779a5fc57e
Revises: 7fcaf5b6e9e8
Create Date: 2017-09-15 17:23:45.307309

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '4e779a5fc57e'
down_revision = '7fcaf5b6e9e8'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('result') as batch_op:
        batch_op.add_column(
            sa.Column('is_unregistered', sa.Boolean(), default=False))


def downgrade():
    with op.batch_alter_table('result') as batch_op:
        batch_op.drop_column('is_unregistered')



from chainerui import _version
from chainerui.client.client import init  # NOQA
from chainerui.client.client import log  # NOQA
from chainerui.client.client import log_reporter  # NOQA


__version__ = _version.__version__



"""empty message.

Revision ID: 1104edb51c0d
Revises: 4e779a5fc57e
Create Date: 2017-09-19 17:49:16.820202

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '1104edb51c0d'
down_revision = '4e779a5fc57e'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table('project',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('path_name',
                              sa.String(length=512), nullable=False),
                    sa.Column('name', sa.String(length=512), nullable=True),
                    sa.PrimaryKeyConstraint('id'),
                    sa.UniqueConstraint('path_name')
                    )


def downgrade():
    op.drop_table('project')



import alembic
from alembic import context
from sqlalchemy import engine_from_config
from sqlalchemy import pool


def run_migrations_online(config):
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine and associate a
    connection with the context.

    """
    connectable = engine_from_config(
        config.get_section(config.config_ini_section),
        prefix='sqlalchemy.',
        poolclass=pool.NullPool)

    with connectable.connect() as connection:
        alembic.context.configure(connection=connection)

        with alembic.context.begin_transaction():
            alembic.context.run_migrations()


def main():
    """main."""
    config = context.config
    config.set_main_option("sqlalchemy.url", config.get_main_option('url'))
    run_migrations_online(config)


main()



import datetime
import json

from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy.orm import relationship
from sqlalchemy import String

from chainerui import database
from chainerui.database import db


class Asset(database.BASE):
    __tablename__ = 'asset'

    id = Column(Integer, primary_key=True)
    result_id = Column(Integer, ForeignKey('result.id'))
    summary = Column(String(512))
    content_list = relationship('Bindata', cascade='all, delete-orphan')
    file_modified_at = Column(DateTime, default=datetime.datetime.now())

    def __init__(self, result_id=None, summary=None, file_modified_at=None):
        self.result_id = result_id
        self.summary = json.dumps(summary)
        self.file_modified_at = file_modified_at
        self.content_list = []

    @classmethod
    def create(cls, result_id=None, summary=None, file_modified_at=None):
        """Initialize an instance and save it to db."""
        asset = cls(result_id, summary, file_modified_at)

        db.session.add(asset)
        db.session.commit()

        return asset

    def __repr__(self):
        return '<Asset id: %r, />' % (self.id)

    @property
    def serialize(self):
        summary = json.loads(self.summary)
        return {
            'train_info': summary,
            'contents': [c.serialize for c in self.content_list]
        }



# -*- coding: utf-8 -*-
import os
from setuptools.command.sdist import sdist
from setuptools import find_packages
from setuptools import setup
import subprocess


with open('requirements.txt') as f:
    required = f.read().splitlines()


with open(os.path.join('docs', 'requirements.txt')) as f:
    docs_required = f.read().splitlines()


requirements = {
    'install': required,
    'docs': docs_required,
    'stylecheck': [
        'autopep8',
        'hacking',
    ],
    'test': [
        'numpy',
        'pytest>=4.0.0,<5.0.0',
        'mock',
    ],
    'test-ci-plain': [
        '-r stylecheck',
        '-r test',
        'pytest-cov',
        'coveralls',
    ],
    'test-ci': [
        '-r test-ci-plain',
        'Pillow',
        'matplotlib',
        'scipy',
        'chainer',
    ]
}


def reduce_requirements(key):
    # Resolve recursive requirements notation (-r)
    reqs = requirements[key]
    resolved_reqs = []
    for req in reqs:
        if req.startswith('-r'):
            depend_key = req[2:].lstrip()
            reduce_requirements(depend_key)
            resolved_reqs += requirements[depend_key]
        else:
            resolved_reqs.append(req)
    requirements[key] = resolved_reqs


for k in requirements.keys():
    reduce_requirements(k)


here = os.path.abspath(os.path.dirname(__file__))
# Get __version__ variable
exec(open(os.path.join(here, 'chainerui', '_version.py')).read())


class chainerui_sdist(sdist):
    def run(self):
        subprocess.check_call('cd frontend && npm run build', shell=True)
        sdist.run(self)


setup(
    name='chainerui',
    version=__version__,  # NOQA
    description='ChainerUI: User Interface for Chainer',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=requirements['install'],
    tests_require=requirements['test'],
    extras_require={k: v for k, v in requirements.items() if k != 'install'},
    package_data={
        'chainerui': [
            'templates/*',
            'static/**/*'
        ],
    },
    author='',
    author_email='',
    url='https://github.com/chainer/chainerui',
    packages=find_packages(exclude=('tests', 'docs')),
    entry_points={
        "console_scripts": [
            "chainerui=chainerui.app:main",
        ]
    },
    cmdclass={
        'sdist': chainerui_sdist
    },
)



import json

from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String

from chainerui import database


class Command(database.BASE):
    """Command Model."""
    __tablename__ = 'command'

    id = Column(Integer, primary_key=True)
    result_id = Column(Integer, ForeignKey('result.id'))
    name = Column(String(1024))
    request = Column(String(1024))
    response = Column(String(1024))

    def __init__(self, name=None, request=None, response=None):
        self.name = name
        self.request = json.dumps(request, indent=4)
        self.response = json.dumps(response, indent=4)

    def update(self, response=None):
        self.response = json.dumps(response, indent=4)

    def __repr__(self):
        return '<Command id: %r />' % (self.id)

    @property
    def serialize(self):
        """serialize."""

        if self.request is None:
            request = None
        else:
            request = json.loads(self.request)

        if self.response is None:
            response = None
        else:
            response = json.loads(self.response)

        return {
            'id': self.id,
            'name': self.name,
            'request': request,
            'response': response
        }



from flask import jsonify
from flask import request
from flask.views import MethodView

from chainerui.database import db
from chainerui.models.project import Project


class ProjectAPI(MethodView):
    """ProjectAPI."""

    def get(self, id=None):
        """get."""

        if id is None:
            path = request.args.get('path_name', default=None)
            if path is not None:
                project = db.session.query(Project).filter_by(
                    path_name=path).first()
                if project is None:
                    return jsonify({
                        'project': None,
                        'message': 'Project path \'%s\' is not found' % path
                    }), 400
                return jsonify({'project': project.serialize})

            projects = db.session.query(Project).all()
            return jsonify({
                'projects': [p.serialize for p in projects]
            })

        else:
            project = db.session.query(Project).filter_by(id=id).first()
            if project is None:
                return jsonify({
                    'project': None,
                    'message': 'No interface defined for URL.'
                }), 404
            return jsonify({
                'project': project.serialize
            })

    def post(self):
        data = request.get_json()  # if invalid data, raise BadRequest
        project_json = data.get('project')
        path = project_json.get('pathName', '')
        if path == '':
            # follow backward compatibility
            path = project_json.get('path_name', '')
        if path == '':
            return jsonify({
                'project': None,
                'message': 'Path of the project is not set.'
            }), 400
        name = project_json.get('name', '')
        if name == '':
            name = path
        crawlable = project_json.get('crawlable', True)

        project = db.session.query(Project).filter_by(path_name=path).first()
        if project is None:
            project = Project.create(
                path_name=path, name=name, crawlable=crawlable)
            return jsonify({
                'project': project.serialize
            })
        else:
            return jsonify({
                'project': None,
                'message': 'Pathname already registered.'
            }), 400

    def put(self, id):
        """put."""

        project = db.session.query(Project).filter_by(id=id).first()

        if project is None:
            return jsonify({
                'project': None,
                'message': 'No interface defined for URL.'
            }), 404

        request_project = request.get_json().get('project')
        project_name = request_project.get('name', None)

        if project_name is not None:
            project.name = project_name

        db.session.add(project)
        db.session.commit()

        return jsonify({
            'project': project.serialize
        })

    def delete(self, id):
        """delete."""
        project = db.session.query(Project).filter_by(id=id).first()

        if project is None:
            response = jsonify({
                'projects': None,
                'message': 'No interface defined for URL.'
            })
            return response, 404

        db.session.delete(project)
        db.session.commit()

        return jsonify({
            'project': project.serialize
        })



import errno
import os

from alembic.command import downgrade
from alembic.command import upgrade
from alembic.config import Config
from sqlalchemy import create_engine
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm import sessionmaker


BASE = declarative_base()
_PACKAGE_DIR = os.path.abspath(os.path.dirname(__file__))


class Database(object):

    def __init__(self):
        self._url = None
        self._initialized = False
        self._engine = None
        self._session = None
        self._external_db = False

        self._alembic_config = None
        self._sqlite_db_file_path = None

    def init_db(self):
        db_dir = self._sqlite_default_db_dir()
        try:
            os.makedirs(db_dir)
        except OSError as e:
            if e.errno == errno.EEXIST and os.path.isdir(db_dir):
                pass
            else:
                raise
        print('DB file path:', db_dir)

    def setup(self, url=None, echo=False):
        if url is None:
            db_dir = self._sqlite_default_db_dir()
            db_path = os.path.join(db_dir, 'chainerui.db')
            self._sqlite_db_file_path = db_path
            url = 'sqlite:///' + db_path
        else:
            self._external_db = True

        if url.startswith('sqlite'):
            connect_args = {'check_same_thread': False}
        else:
            connect_args = {}
        self._url = url

        try:
            engine = create_engine(
                url,
                connect_args=connect_args,
                echo=echo
            )
        except ImportError as e:
            print(e, ', Please install the driver to support the external DB')
            return False
        self._engine = engine

        if not self._check():
            return False

        self._session = scoped_session(
            sessionmaker(autocommit=False, autoflush=False, bind=engine)
        )
        self._initialized = True

        self._setup_alembic_config(url)

        return True

    def _setup_alembic_config(self, url):
        ini_path = os.path.join(_PACKAGE_DIR, 'alembic.ini')
        config = Config(ini_path)
        config.set_main_option(
            'script_location', os.path.join(_PACKAGE_DIR, 'migration'))
        config.set_main_option('url', url)
        self._alembic_config = config

    def _check(self):
        if not self._external_db:
            if not os.path.isdir(self._sqlite_default_db_dir()):
                print('DB is not initialized, please run '
                      '\'chainerui db create\' command before')
                return False
        try:
            connection = self._engine.connect()
            connection.close()
        except OperationalError as e:
            print(e)
            return False
        return True

    def upgrade(self):
        upgrade(self.alembic_config, 'head')

    def downgrade(self):
        downgrade(self.alembic_config, 'base')

    def remove_db(self):
        if not self._external_db and self._sqlite_db_file_path is not None:
            if os.path.exists(self._sqlite_db_file_path):
                os.remove(self._sqlite_db_file_path)
        self.__init__()  # initialize all attribute

    def _sqlite_default_db_dir(self):
        root = os.path.abspath(
            os.path.expanduser(os.getenv('CHAINERUI_ROOT', '~/.chainerui')))
        return os.path.join(root, 'db')

    @property
    def url(self):
        if not self._initialized:
            raise ValueError('not setup DB URL')
        return self._url

    @property
    def engine(self):
        if not self._initialized:
            raise ValueError('not setup database engine')
        return self._engine

    @property
    def session(self):
        if not self._initialized:
            raise ValueError('not setup database session')
        return self._session

    @property
    def alembic_config(self):
        if not self._initialized:
            raise ValueError('not setup migration configuration')
        return self._alembic_config


db = Database()



import json
import os
import shutil
import time

from chainerui.utils.save_args import save_args
from chainerui.utils.tempdir import tempdir


# Select the best-resolution timer function
# copied from chainer.training.trainer
try:
    _get_time = time.perf_counter
except AttributeError:
    if os.name == 'nt':
        _get_time = time.clock
    else:
        _get_time = time.time


class LogReport(object):

    """Util class to output 'log' file.

    This class supports to output 'log' file. The file spec follows
    :class:`chainer.extensions.LogReport`, however, 'epoch' and 'iteration'
    are not set automatically, and need to set these values.

    Args:
        out_path (str): Output directory name to save conditions.
        conditions (:class:`argparse.Namespace` or dict): Experiment conditions
            to show on a job table. Keys are show as table header and values
            are show at a job row.

    """

    def __init__(self, out_path, conditions=None):
        self._out_path = out_path
        self._log_name = 'log'
        self._log = []
        self._start_at = _get_time()

        if conditions is not None:
            # if not exist the path, `save_args` makes it
            save_args(conditions, out_path)
        else:
            try:
                os.makedirs(out_path)
            except OSError:
                pass

    def __call__(self, stats):
        """Add training log.

        Args:
            stats (dict): Training log values. The object must be key-value
                style and values type must be `float` or `int`. When the object
                does not have 'elapsed_time' key, the function set the time
                automatically. The measurement starts when create new instance.
        """
        if 'elapsed_time' not in stats:
            stats['elapsed_time'] = _get_time() - self._start_at
        self._log.append(stats)

        with tempdir(prefix=self._log_name, dir=self._out_path) as tempd:
            path = os.path.join(tempd, 'log.json')
            with open(path, 'w') as f:
                json.dump(self._log, f, indent=4)

            new_path = os.path.join(self._out_path, self._log_name)
            shutil.move(path, new_path)



from __future__ import absolute_import

import logging
import threading

import structlog


_mutex = threading.Lock()
_logger = None


def _get_library_logger():
    return logging.getLogger(__name__)


def set_loglevel(level):
    _get_library_logger().setLevel(level)


def get_logger():
    global _logger

    with _mutex:
        if _logger is not None:
            return _logger
        structlog.configure(
            processors=[
                structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
            ],
            logger_factory=structlog.stdlib.LoggerFactory(),
        )

        own_logger = _get_library_logger()
        if not logging.getLogger().handlers:
            # own hander is set only when python root logger is not setup
            formatter = structlog.stdlib.ProcessorFormatter(
                processor=structlog.dev.ConsoleRenderer(colors=False),
            )
            handler = logging.StreamHandler()
            handler.setFormatter(formatter)
            own_logger.addHandler(handler)

        own_logger.setLevel(logging.INFO)
        _logger = structlog.get_logger(__name__)
    return _logger


logger = get_logger()



import datetime
import os
import time

from chainerui.client.helper import urlopen
from chainerui.utils.save_args import convert_dict


def get_default_url():
    return os.getenv('CHAINERUI_URL', 'http://localhost:5000')


def get_default_project_path():
    return os.getenv('CHAINERUI_PROJECT_PATH', os.path.abspath(os.getcwd()))


def make_timestamp(dt=None):
    if dt is None:
        dt = datetime.datetime.now()
    return time.mktime(dt.timetuple()) + dt.microsecond / 1e6


def format_datetime(dt=None):
    if dt is None:
        dt = datetime.datetime.now()
    return dt.strftime('%Y%m%d-%H%M%S.%f')


class Client(object):

    def __init__(self, url=None, crawlable=False):
        if url is None:
            url = get_default_url()
        self.url = (url[:len(url)-1] if url.endswith('/') else url) + '/api/v1'
        self.crawlable = crawlable

        # Basically, the project path should be set current working directory
        # to menage experiments easier, but on some environment, working
        # directories are same at all times. To handle this, use environment
        # variable.
        self.project_path = get_default_project_path()
        self.project_id = None
        self.result_id = None
        self.cached_logs = []
        self.first_reset = False

    def setup_project(self, project_name=None):
        """Setup project on ChainerUI server

        Project path on default is working directory. First, check the project
        path is registered or not, and if not found, register the project path.
        Second, if ``project_name`` is set and changed, update the project
        name.

        Args:
            project_name (str): If set, update the project name. If set
                ``None`` (on default), try to get environment variable,
                ``'CHAINERUI_PROJECT_NAME'`` if set.

        Returns:
            bool: When succeed to setup, return ``True``
        """

        project_path = self.project_path
        if project_name is None:
            project_name = os.getenv('CHAINERUI_PROJECT_NAME', None)

        # check the path exists or not
        check_url = '{}?path_name={}'.format(self.projects_url, project_path)
        project_res, msg = urlopen('GET', check_url)
        # if project is not found, create a project
        if project_res is None:
            if (project_path + '\' is not found') not in msg:
                print('connection error, URL: {}, Message: {}'.format(
                    check_url, msg))
                return False
            # if message includes pathname, means not found error
            # continue to process to register this project
            project_req = {
                'project': {
                    'path_name': project_path,
                    'name': project_name,
                    'crawlable': self.crawlable,
                },
            }
            project_res, msg = urlopen(
                'POST', self.projects_url, data=project_req)
            if project_res is None:
                print('register error, URL: {}, Message: {}'. format(
                    self.projects_url, msg))
                return False

        project_id = project_res['project']['id']
        if project_name is not None and\
                project_res['project']['name'] != project_name:
            # update project name
            project_req = {'project': {'name': project_name}}
            project_res, msg = urlopen(
                'PUT', self.project_url, data=project_req)
            if project_res is None:
                print('update error, URL: {}, Message: {}'.format(
                    self.project_url, msg))
                # fail to update project name, but the target project has been
                # registered, so not return False

        self.project_id = project_id
        return True

    def register_result(self, result_name=None, overwrite_result=False):
        """Register result on ChaienrUI server

        Basically result path is same as project path, but to be unique key,
        add start time on the result path. The result path is virtual one,
        not make physical directory. If ``overwrite_result`` set ``True``,
        the client tool not add start time and continue to use same result
        record.

        Args:
            result_name (str): If set, update the result name. If set
                ``None`` (on default), try to get environment variable,
                ``'CHAINERUI_RESULT_NAME'`` if set.
            overwrite_result (bool): If set ``True``, not set start time on
                the result path and overwrite data on the result.

        Returns:
            bool:  When succeed to setup, return ``True``
        """

        now = datetime.datetime.now()
        result_path = self.project_path
        if result_name is None:
            result_name = os.getenv('CHAINERUI_RESULT_NAME', None)

        if overwrite_result:
            check_url = '{}?path_name={}'.format(
                self.results_url, result_path)
            result_res, msg = urlopen('GET', check_url)
            if result_res is None:
                if (result_path + '\' is not found') not in msg:
                    print('connection error, URL: {}, Message: {}'.format(
                        check_url, msg))
                    return False
            else:
                # result is found, overwrite on the result record
                self.result_id = result_res['result']['id']
                self.first_reset = True
                if result_name is not None and\
                        result_res['result']['name'] != result_name:
                    # update result name
                    result_req = {'result': {'name': result_name}}
                    result_res, msg = urlopen(
                        'PUT', self.result_url, data=result_req)
                    if result_res is None:
                        print('update error, URL: {}, Message: {}'.format(
                            self.result_url, msg))
                        # fail to update result name, but the target result
                        # has been found, so not return False
                return True
        else:
            # result path is required unique key, make dummy path but not
            # make the physical directory.
            result_path += '-' + format_datetime(now)
            if result_name is not None:
                result_name += '-' + format_datetime(now)

        result_req = {
            'result': {
                'pathName': result_path,
                'name': result_name,
                'crawlable': False,
                'logModifiedAt': make_timestamp(now),
            }
        }
        result_res, msg = urlopen('POST', self.results_url, data=result_req)
        if result_res is None:
            print('register error, URL: {}, Message: {}'.format(
                self.results_url, msg))
            return False
        self.result_id = result_res['result']['id']
        return True

    def post_log(self, logs, modified_at=None):
        """Post logs

        If fail to post the ``logs``, the data is cached and try to post them
        next time with new one.

        Args:
            logs (list): stats data to post.
            modified_at (datetime): last modified time.
        """

        if modified_at is None:
            modified_at = make_timestamp()
        self.cached_logs.extend(logs)
        log_req = {
            'log': {
                'values': self.cached_logs,
                'modifiedAt': modified_at,
                'reset': self.first_reset,
            }
        }
        log_res, msg = urlopen('POST', self.logs_url, data=log_req)
        if log_res is None:
            print('post log error, URL: {}, Message: {}'.format(
                self.logs_url, msg))
            return

        if log_res['logs']['totalLogCount'] >= 0:
            self.first_reset = False
            self.cached_logs = []

    def post_args(self, conditions):
        arg_req = {'argument': conditions}
        res, msg = urlopen('POST', self.args_url, data=arg_req)
        if res is None:
            print('post args error, URL: {}, Message: {}'.format(
                self.args_url, msg))
            return False
        return True

    @property
    def projects_url(self):
        return '{}/projects'.format(self.url)

    @property
    def project_url(self):
        return '{}/{}'.format(self.projects_url, self.project_id)

    @property
    def results_url(self):
        return '{}/results'.format(self.project_url)

    @property
    def result_url(self):
        return '{}/{}'.format(self.results_url, self.result_id)

    @property
    def logs_url(self):
        return '{}/logs'.format(self.result_url)

    @property
    def args_url(self):
        return '{}/args'.format(self.result_url)


_client = None


def init(url=None, project_name=None, result_name=None, overwrite_result=False,
         crawlable=False, conditions=None):
    """Initialize client tools

    Initialize client object, then setup project and result. Even if some
    errors are occurred, client object is set ``None`` and without exception.

    Args:
        url (str): ChainerUI server URL. set ``'localhost:5000'`` on default.
        project_name (str): project name is set from project path, working
            directory on default. If set, ChainerUI shows the name instead
            the project path.
        result_name (str): result name is set project path + start time
            on default. If set, ChainerUI shows the name + start time instead
            the path.
        overwrite_result (bool): the client tool make different job results
            every time on default. If set ``True``, the client tool posts
            logs on the same result.
        crawlable (bool): to inform server not to crawl physical logs.
        conditions (:class:`argparse.Namespace` or dict): Experiment conditions
            to show on a job table. Keys are show as table header and values
            are show at a job row.
    """

    global _client
    if _client is None:
        _client = Client(url=url, crawlable=crawlable)

    if not _client.setup_project(project_name=project_name):
        _client = None
        return

    if not _client.register_result(
            result_name=result_name, overwrite_result=overwrite_result):
        _client = None
        return

    if conditions is not None:
        args = convert_dict(conditions)
        _client.post_args(args)


def log_reporter():
    """Log reporter via POST API

    Return a callback function to post a log to a ChainerUI server. If the
    initialization (see ``chainerui.init()``) is failed, the callback function
    do nothing when called. If the initialization is done but fail to send the
    log by some kind of error, the log is cached and try to post it next time
    with new one.

    The callback function is expected to use with ``postprocess``
    option of Chainer ``LogReport`` extension::

       >>> chainerui.init()
       >>>
       >>> trainer = chainer.training.Trainer(...)
       >>> trainer.extend(
       >>>     extensions.LogReport(postprocess=chainerui.log_reporter()))

    Returns:
        func: function.
    """
    class PostProcess(object):

        def __call__(self, stats_cpu):
            if _client is None:
                return

            _client.post_log([stats_cpu])

    return PostProcess()


def log(value):
    """Send log

    Send log data and will be shown ad training log on web browser.

    Example::

       >>> chainerui.init()
       >>> # ...
       >>> stats = {
       >>>     'epoch': epoch,
       >>>     'train/loss': loss,
       >>>     }
       >>> chainerui.log(stats)

    Args:
        value (dict): target log.
    """
    if _client is None:
        return
    _client.post_log([value])
